*************************************************************************************
*														*
*	Rain impact simulation for the EASy68k simulator	2006/11/15			*
*														*
*	This code uses double buffering to render a patterm similar to that made	*
*	by raindrops landing in shallow water on a tiled floor. It's all a big fudge	*
*	and a bit slow on the bigger arrays, but makes some pretty patterns and		*
*	shows how such effects can be made.								*
*														*
*	This version for Sim68K 3.2 or later							*
*														*
*	More 68000 and other projects can be found on my website at ..			*
*														*
*	http://www.themotionstore.com/leeedavison/index.html					*
*														*
*	.. or ..												*
*														*
*	http://members.lycos.co.uk/leeedavison/index.html					*
*														*
*	mail : leeedavison@lycos.co.uk								*
*														*
*************************************************************************************

size	EQU	$40					* size, both x and y, of the environment.
							* $80 is the recommended maximum as beyond
							* that you end up with -ve offsets later and
							* you'd need to change the MOVEQs
size2	EQU	size*size-1				* maximum array index
mask	EQU	$07					* mask for new drop generation. chance is
							* 1-(n/(n+1)). for heavier rain make this $03

	ORG	$1000

start
	BSR		Initialise			* go setup stuff

	MOVE.b	#17,d1			* enable double buffering
	MOVE.b	#92,d0			* set draw mode
	TRAP		#15

	MOVE.b	#11,d0			* position cursor
	MOVE.w	#$FF00,d1			* clear screen
	TRAP		#15

loop
	MOVE.b	#94,d0			* copy screen buffer to main
	TRAP		#15

	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	MOVE.l	d1,-(sp)			* push time on the stack, erk! needed d7

	MOVE.b	#11,d0			* position cursor
	MOVE.w	#$FF00,d1			* clear screen buffer
	TRAP		#15

	BSR.s		gen_array			* generate and draw the new screen data

* now wait for a while. if you have a slow machine this shouldn't slow it down much but
* if you have a fast machine them this will stop the effect running too fast

	MOVE.l	(sp)+,d7			* restore time in 1/100 ths from stack
wait_100ms
	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	SUB.l		d7,d1				* subtract previous time from current time
	CMP.b		#$09,d1			* compare with 9/100ths
	BMI.s		wait_100ms			* loop if time not up yet

* if you want to see how fast it can go comment out the BMI.s above

	BRA.s		loop				* loop forever

*************************************************************************************
*
* generate the next stage and display it

gen_array
	BSR		LAB_RND			* call the PRNG code
	MOVEQ		#mask,d1			* set size mask
	AND.b		PRNlword(a3),d1		* mask byte from the result
	BNE.s		no_raindrop			* if not time for another drop skip new

	BSR		LAB_RND			* call the PRNG code
	MOVE.w	PRNlword(a3),d1		* get word from result, select any array cell
	AND.w		#size2,d1			* mask to array bounds
	MOVE.b	#$FE,(a1,d1)		* set new raindrop

* for each pixel new pixel = ABS(sum of the eight surrounding pixels / 4 - previous value)
* this is a big fudge, any more and it feeds back and just makes noise, any less and the
* drops have too short a life. there is no decay value for the waves, that is taken care
* of by the loss to rounding on each step

no_raindrop
	MOVEQ		#size-1,d7			* set y index

* lots of time is saved by making use of the registers. for each y loop the y-1, y and
* y+1 base addresses are calculated for array 1 as well as the y base address for array
* 2. this means only one MULU for array acess is needed, the rest is done with ADDs

gen_yloop
	MOVE.w	#size,d6			* set x index
	MOVE.w	d7,d5				* copy y index
	MULU.w	d6,d5				* calculate y offset
	LEA		(a2,d5),a0			* y base address, array 2
	LEA		(a1,d5),a5			* y base address

	MOVE.w	d5,d0				* copy y offset

	SUB.w		d6,d5				* calculate y - 1 offset
	AND.w		#size2,d5			* mask to array size
	LEA		(a1,d5),a4			* y - 1 base address

	ADD.w		d6,d0				* calculate y + 1 offset
	AND.w		#size2,d0			* mask to array bounds
	LEA		(a1,d0),a6			* y + 1 base address

	SUBQ.w	#1,d6				* set x index

* we have y-1, y and y+1 in a4, a5 and a6, now for each x loop first calculate the x-1 and
* x+1 values and use these as offsets to the y base addresses to access the values of the
* eight adjoining pixels to the current pixel

gen_xloop
	MOVEQ		#size-1,d0			* clear high three bytes of total, the lower
							* byte gets overwritten
	MOVEQ		#0,d2				* clear longword

	MOVEQ		#1,d5				* set + 1
	ADD.w		d6,d5				* add x index
	AND.w		d0,d5				* mask to array size

	MOVEQ		#-1,d4			* set - 1
	ADD.w		d6,d4				* add x index
	AND.w		d0,d4				* mask to array size

* now we have x-1, x and x+1 in d4, d5 and d6. using these with a4, a5 and a6 allows easy
* access to the pixel elements

	MOVE.b	(a4,d4),d0			* get pixel intensity (-1, -1)
	MOVE.b	(a4,d6),d2			* get pixel intensity ( 0, -1)
	ADD.w		d2,d0				* add to total
	MOVE.b	(a4,d5),d2			* get pixel intensity (+1, -1)
	ADD.w		d2,d0				* add to total

	MOVE.b	(a5,d4),d2			* get pixel intensity (-1,  0)
	ADD.w		d2,d0				* add to total
	MOVE.b	(a5,d5),d2			* get pixel intensity (+1,  0)
	ADD.w		d2,d0				* add to total

	MOVE.b	(a6,d4),d2			* get pixel intensity (-1, +1)
	ADD.w		d2,d0				* add to total
	MOVE.b	(a6,d6),d2			* get pixel intensity ( 0, +1)
	ADD.w		d2,d0				* add to total
	MOVE.b	(a6,d5),d2			* get pixel intensity (+1, +1)
	ADD.w		d2,d0				* add to total

	ASR.w		#2,d0				* surrounding pixel total \ 4

	MOVE.b	(a0,d6),d2			* get previous pixel intensity
	SUB.w		d2,d0				* subtract from surrounding pixel total \ 4
	BPL.s		pos_pixel			* branch if value +ve

	NEG.w		d0				* else use ABS(value)
pos_pixel
	MOVE.b	d0,(a0,d6)			* save new pixel intensity

* the following five lines add a formula texture to the background, they can be commented
* out for a black background

	MOVE.b	d7,d1				* get y
	EOR.b		d6,d1				* add x
	ASL.b		#1,d1				* * 2, make the small tiles 2 x 2 pixels
	ADD.b		#$20,d1			* set offset from black
	EOR.b		d1,d0				* modify the intensity with the texture

* now set the intensity for this (modified) pixel as the draw colour and the fill colour

	MOVEQ		#0,d1				* clear colour longword
	MOVE.b	d0,d1				* copy colour byte
	SWAP		d1				* shift to high word
	MOVE.b	d0,d1				* copy colour byte - again
	ASL.w		#8,d1				* shift to word high byte
	OR.b		d0,d1				* copy colour byte - yet again
	MOVE.b	#80,d0			* set pen colour
	TRAP		#15
	MOVE.b	#81,d0			* set fill colour
	TRAP		#15

* now draw a small 4 x 4 square of pixels

	MOVEQ		#$08,d1			* x offset / 4
	ADD.w		d6,d1				* + x co-ordinate
	ASL.w		#2,d1				* * 4

	MOVE.w	#$08,d2			* y offset / 4
	ADD.w		d7,d2				* + y co-ordinate
	ASL.w		#2,d2				* * 4

	MOVE.w	d1,d3				* copy x co-ordinate
	ADDQ.w	#3,d3				* make a small square

	MOVE.w	d2,d4				* copy y co-ordinate
	ADDQ.w	#3,d4				* make a small square

	MOVEQ		#87,d0			* draw filled rectangle
	TRAP		#15				* do draw

	DBF		d6,gen_xloop		* decrement x and loop for next

	DBF		d7,gen_yloop		* decrement y and loop for next

	EXG		a1,a2				* swap the current and last arrays
	RTS

*************************************************************************************
*
* setup stuff. set the pointers, initialise the PRNG seed and clear the arrays


Initialise
	LEA		(LAB_VARS,PC),a3		* get pointer to variables
	LEA		array1(a3),a1		* get pointer to array 1
	LEA		array2(a3),a2		* get pointer to array 2
	MOVE.l	#$31415927,PRNlword(a3)	* set PRNG seed. this can be any value
							* except zero, changing this will produce
							* a different start point for the effect
	MOVEQ		#0,d0				* clear d0
	MOVE.l	#size*size/2-1,d1		* get array size/2
	MOVEA.l	a1,a4				* copy array memory start
clear_loop
	MOVE.l	d0,(a4)+			* clear longword
	DBF		d1,clear_loop		* decrement count and loop for next

	RTS

*************************************************************************************
*
* This is the code that generates the pseudo random sequence. A seed word located in
* PRNlword(a3) is loaded into a register before being operated on to generate the
* next number in the sequence. This number is then saved as the seed for the next
* time it's called.
*
* This code is adapted from the 32 bit version of RND(n) used in EhBASIC68. Taking
* the 19th next number is slower but helps to hide the shift and add nature of this
* generator as can be seen from analysing the output.

LAB_RND
	MOVEM.l	d0-d2,-(sp)			* save d0, d1 and d2
	MOVE.l	PRNlword(a3),d0		* get current seed longword
	MOVEQ		#$AF-$100,d1		* set EOR value
	MOVEQ		#18,d2			* do this 19 times
Ninc0
	ADD.l		d0,d0				* shift left 1 bit
	BCC.s		Ninc1				* if bit not set skip feedback

	EOR.b		d1,d0				* do Galois LFSR feedback
Ninc1
	DBF		d2,Ninc0			* loop

	MOVE.l	d0,PRNlword(a3)		* save back to seed longword
	MOVEM.l	(sp)+,d0-d2			* restore d0, d1 and d2

	RTS

*************************************************************************************
*
* variables used

LAB_VARS
							* points to variables base
	OFFSET	0				* going to use relative addressing

PRNlword	ds.l	1				* PRNG seed long word

array1	ds.b	size*size			* make space for the first array
array2	ds.b	size*size			* make space for the second array

	ORG	*

	END	start
